'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});


const {
  GraphQLEnumType,
  GraphQLInputType,
  GraphQLInputObjectType,
  GraphQLInterfaceType,
  GraphQLList,
  GraphQLNonNull,
  GraphQLObjectType,
  GraphQLScalarType,
  GraphQLType,
  GraphQLUnionType
} = require('graphql');

// const t = require('babel-types');
// import {
//   Program,
//   ExportNamedDeclaration, 
//   Node,
//   ObjectTypeAnnotation,
//   ObjectTypeProperty,
//   Identifier,
//   NullableTypeAnnotation,
//   GenericTypeAnnotation,
//   TypeParameterInstantiation,
// } from 'babel-types';

const {
  SchemaUtils
} = require('relay-compiler/lib/GraphQLCompilerPublic');

class ScalaGen {
  constructor(ast) {
    this.root = ast;
    this._toPrint = new Map();
    this._seenTypes = new Set();
    this._curClass = [];
  }

  newClass(s, n) {
    let newCls = s;
    if (n) {
      newCls = `${newCls}_${n}`;
    }
    const result = this._curClass.find(cls => cls == newCls);
    if (result) {
      const num = n ? n + 1 : 1;
      return this.newClass(s, num);
    } else {
      this._curClass.unshift(newCls);
      return newCls;
    }
  }

  popClass() {
    return this._curClass.shift();
  }

  printClass() {
    this.print(this._curClass[0]);
  }

  print(s) {
    const cls = this._curClass[0];
    const v = this._toPrint.get(cls);
    if (v) {
      v.push(s);
    } else {
      this._toPrint.set(cls, [s]);
    }
  }

  ret() {
    this.print("\n");
  }

  openB() {
    this.print("{\n");
  }
  closeB() {
    this.print("\n};\n");
  }

  start() {
    this.run(this.root);
  }

  jsPrefix() {
    return "js";
  }

  anyType() {
    this.print(`${this.jsPrefix()}.Any`);
  }

  anyObjType() {
    this.print(`${this.jsPrefix()}.Dictionary[${this.jsPrefix()}.Any]`);
  }

  valDefName(name) {
    this.print("val");
    this.print(name);
    this.print(":");
  }

  valDef(name, tpe) {
    this.valDefName(name);
    this.print(tpe);
    this.ret();
  }

  traitDefOpen(ex) {
    this.print("@ScalaJSDefined");
    this.print("\ntrait");
    this.printClass();
    this.print("extends");
    this.print(`${this.jsPrefix()}.Object`);
    if (ex) {
      this.print("with");
      this.print(ex);
    }
    this.openB();
  }

  concatParents(p, p2) {
    return p + p2;
  }

  titleCase(s) {
    return s.charAt(0).toUpperCase() + s.substr(1);
  }

  transformNonNullableScalarType(type, backupType) {
    if (type instanceof GraphQLList) {
      // throw new Error(`Could not convert from GraphQL type ${type.toString()}`);
      this.print("js.Array[");
      this.transformScalarType(type.ofType, backupType);
      this.print("]");
      return;
    } else if (type instanceof GraphQLObjectType || type instanceof GraphQLUnionType || type instanceof GraphQLInterfaceType) {
      // throw new Error(`Could not convert from GraphQL type ${type.toString()}`);
      if (!backupType) {
        throw new Error(`Could not convert from GraphQL type ${type.toString()}, missing backup type`);
      }
      this.print(backupType);
    } else if (type instanceof GraphQLScalarType) {
      return this.transformGraphQLScalarType(type, backupType);
    } else if (type instanceof GraphQLEnumType) {
      // TODO: String for now.
      return this.print("String");
    } else {
      throw new Error(`Could not convert from GraphQL type ${type.toString()}`);
    }
  }

  transformGraphQLScalarType(type, backupType) {
    switch (type.name) {
      case 'ID':
      case 'String':
      case 'Url':
        return this.print("String");

      case 'Float':
      case 'Int':
      case 'BigDecimal':
      case 'BigInt':
      case 'Long':
        return this.print("Double");
      case 'Boolean':
        return this.print("Boolean");
      default:
        if (backupType) {
          this.print(backupType);
        } else {
          this.print(`${this.jsPrefix()}.Any`);
        }
    }
  }

  transformScalarType(type, backupType) {
    if (type instanceof GraphQLNonNull) {
      return this.transformNonNullableScalarType(type.ofType, backupType);
    } else {
      // $FlowFixMe
      return this.transformNonNullableScalarType(type, backupType);
    }
  }

  handleSelections(parent, a) {
    const onlyClasses = a.selections.every(s => s.kind === "FragmentSpread");
    const onlyMembers = a.selections.every(s => s.kind !== "FragmentSpread");

    let valName = a.name;
    let tpe;

    if (onlyClasses) {
      // $FlowFixMe
      const clss = a.selections.map(s => s.name);
      tpe = clss.join(" with ");
    } else {
      // We're going to have to make a new class, this is the name.
      tpe = this.newClass(parent.join('') + this.titleCase(a.name));

      // Get all the mixed in Fragments to Inherit from.
      const parentTpe = a.selections.filter(s => s.kind === "FragmentSpread")
      // $FlowFixMe
      .map(s => s.name);

      const newParentTpe = parentTpe.join(" with ");
      // create a new class here.
      this.traitDefOpen(newParentTpe);
      a.selections.filter(s => s.kind !== "FragmentSpread").map(s => this.runSelection(parent.concat([this.titleCase(a.name)]), s));

      this.closeB();
      this.ret();
      // Completed that class
      this.popClass();
    }

    this.valDefName(valName);
    // $FlowFixMe
    this.transformScalarType(a.type, tpe);
    this.ret();
  }

  runSelection(parent, a) {
    switch (a.kind) {
      case "ScalarField":
        const csf = a;
        this.valDefName(csf.name);
        if (csf.type) {
          let required = false;
          // $FlowFixMe
          this.transformScalarType(csf.type);
        } else {
          `${this.jsPrefix()}.Any`;
        }
        this.ret();
        break;

      case "LinkedField":
        const clf = a;
        // parent.push(this.titleCase(a.name));
        this.handleSelections(parent, clf);
        break;

      case "FragmentSpread":
        console.error("Shouldn't encounter a FragmentSpread directly. ", a, parent);
        break;
    }
  }

  run(a) {
    switch (a.kind) {
      case 'Fragment':
        const p = a;
        this.newClass(p.name);

        const clss = p.selections.filter(s => s.kind === "FragmentSpread")
        // $FlowFixMe
        .map(s => s.name).join(' with ');

        this.traitDefOpen(clss);
        p.selections.filter(s => s.kind !== "FragmentSpread").map(s => this.runSelection([p.name], s));
        this.closeB();

        break;
      case 'Root':
        const r = a;
        this.newClass(r.name);

        this.traitDefOpen();
        r.selections.map(s => this.runSelection([r.name], s));
        this.closeB();
      // case "ExportNamedDeclaration":
      //   const end: ExportNamedDeclaration = a;

      //   if (end.declaration.type == "TypeAlias") {
      //     this.print("trait");
      //   }

      //   this.run(end.declaration.id);
      //   this.print("extends");
      //   this.print(`${this.jsPrefix()}.Object`);
      //   this.openB();
      //   this.ret();
      //   this.run(end.declaration.right);
      //   this.closeB();
      //   break;

      // case "ObjectTypeAnnotation":
      //   const ota: ObjectTypeAnnotation = a;
      //   if (ota.properties.length == 0) {
      //     this.anyObjType();
      //   } else {
      //     ota.properties.map((s) => this.run(s))
      //   }
      //   break;

      // case "GenericTypeAnnotation":
      //   const gta: GenericTypeAnnotation = a;
      //   switch (gta.id.name) {
      //     case "$ReadOnlyArray":
      //       this.print(`${this.jsPrefix()}.Array[`);
      //       this.run(gta.typeParameters);
      //       this.print(']');
      //       break;
      //   };
      //   break;

      // case "ObjectTypeProperty":
      //   const otp: ObjectTypeProperty = a;
      //   this.print("val")
      //   this.run(otp.key);
      //   this.print(":");
      //   this.run(otp.value);
      //   this.ret();
      //   break;

      // case "NullableTypeAnnotation":
      //   const nta: NullableTypeAnnotation = a;
      //   this.run(nta.typeAnnotation)
      //   break;

      // case "Identifier":
      //   const identifier: Identifier = a;
      //   this.print(identifier.name)
      //   break;

      // case "StringTypeAnnotation":
      //   this.print("String");
      //   break;

      // case "TypeParameterInstantiation":
      //   const tpi: TypeParameterInstantiation = a;
      //   tpi.params.map((s) => this.run(s));
      //   break;
    }
  }

  out() {
    const arr = [];
    for (const v of this._toPrint.entries()) {
      arr.push(v[1].join(" "));
    }

    return arr.join("\n");
  }
}
exports.ScalaGen = ScalaGen;